import {DynamoDBClient, GetItemCommand, QueryCommand, ScanCommand} from '@aws-sdk/client-dynamodb'
import type {QueryCommandOutput} from '@aws-sdk/client-dynamodb'
import type {AttributeValue} from '@aws-sdk/client-dynamodb'
import {unmarshall} from '@aws-sdk/util-dynamodb'
import type {DDBClientConfiguration} from './DDBClient.ts'
import {formatKey} from './formatter/formatKey.ts'
import {mapGetAllItemsCommand} from './mapper/mapGetAllItemsCommand.ts'
import {mapGetItemCommand} from './mapper/mapGetItemCommand.ts'
import {mapGetItemIndexCommand} from './mapper/mapGetItemIndexCommand.ts'
import {mapGetItemsCommand} from './mapper/mapGetItemsCommand.ts'

export interface DDBReadOnlyClientConfiguration<
  Item extends object,
  PKName extends keyof Item,
  SKName extends keyof Item | never = never,
> extends DDBClientConfiguration<Item, PKName, SKName> {
  indexName?: string
}

export class DDBReadOnlyClient<
  Item extends object,
  PKName extends keyof Item,
  SKName extends keyof Item | never = never,
> {
  protected config: DDBReadOnlyClientConfiguration<Item, PKName, SKName>
  protected client: DynamoDBClient

  constructor(config: DDBReadOnlyClientConfiguration<Item, PKName, SKName>) {
    this.config = config
    this.client = new DynamoDBClient(config.configuration || {})
  }

  async getItem(partitionKey: Item[PKName], sortKey?: Item[SKName]): Promise<Item | null> {
    let command = undefined
    try {
      this.requireFullPrimaryKey(partitionKey, sortKey)

      if (!this.config.indexName) {
        command = mapGetItemCommand(this.config, partitionKey, sortKey)
        const response = await this.client.send(new GetItemCommand(command))

        return response?.Item ? (unmarshall(response?.Item) as Item) : null
      }

      command = mapGetItemIndexCommand(this.config, partitionKey, sortKey)
      const response = await this.client.send(new QueryCommand(command))
      const items = this.extractItems(response)
      return items.length > 0 ? items[0] : null
    } catch (err) {
      console.error(
        'Unable to get item:',
        formatKey(this.config, partitionKey, sortKey),
        'with command',
        JSON.stringify(command),
      )
      throw err
    }
  }

  async getItems(partitionKey: Item[PKName], sortKeyPrefix?: string, pageSize?: number): Promise<Item[]> {
    let command = undefined
    try {
      let items: Item[] = []
      let lastEvaluatedKey: Record<string, AttributeValue> | undefined = undefined
      // eslint-disable-next-line no-constant-condition
      while (true) {
        command = mapGetItemsCommand(this.config, partitionKey, sortKeyPrefix, pageSize, lastEvaluatedKey)
        const response: QueryCommandOutput = await this.client.send(new QueryCommand(command))
        lastEvaluatedKey = response.LastEvaluatedKey
        items = [...items, ...this.extractItems(response)]

        if (!lastEvaluatedKey) break
      }

      return items
    } catch (err) {
      console.error(
        'Unable to get items:',
        formatKey(this.config, partitionKey, sortKeyPrefix),
        'with command',
        JSON.stringify(command),
      )
      throw err
    }
  }

  async getAllItems(pageSize?: number): Promise<Item[]> {
    let command = undefined
    try {
      let items: Item[] = []
      let lastEvaluatedKey: Record<string, AttributeValue> | undefined = undefined
      // eslint-disable-next-line no-constant-condition
      while (true) {
        command = mapGetAllItemsCommand(this.config, pageSize, lastEvaluatedKey)
        const response: QueryCommandOutput = await this.client.send(new ScanCommand(command))
        lastEvaluatedKey = response.LastEvaluatedKey
        items = [...items, ...this.extractItems(response)]

        if (!lastEvaluatedKey) break
      }

      return items
    } catch (err) {
      console.error('Unable to get all items with command', JSON.stringify(command))
      throw err
    }
  }

  private extractItems(response: QueryCommandOutput): Item[] {
    return response?.Items?.map((item) => unmarshall(item) as Item) || []
  }

  protected requireFullPrimaryKey(partitionKey: unknown, sortKey: unknown | undefined): void {
    if (!partitionKey) throw new Error(`Missing partitionKey.`)

    if (this.config.sortKeyName && !sortKey) throw new Error(`Missing sortKey.`)
  }
}
