import type {UpdateItemCommandInput} from '@aws-sdk/client-dynamodb'
import {marshall} from '@aws-sdk/util-dynamodb'
import type {DDBReadOnlyClientConfiguration} from '../DDBReadOnlyClient.ts'
import {replaceUndefinedValuesWithNull} from '../utils/replaceUndefinedValuesWithNull.ts'
import {mapKey} from './mapKey.ts'

export const mapUpdateItemCommand = <
  Item extends object,
  PKName extends keyof Item,
  SKName extends keyof Item | never = never,
>(
  config: DDBReadOnlyClientConfiguration<Item, PKName, SKName>,
  item: Partial<Item>,
  createIfNotExists = false,
): UpdateItemCommandInput => {
  const partitionKeyName = String(config.partitionKeyName)
  const partitionKey = item[config.partitionKeyName]
  const sortKeyName = config.sortKeyName ? String(config.sortKeyName) : undefined
  const sortKey = config.sortKeyName ? item[config.sortKeyName] : undefined

  return {
    TableName: config.tableName,
    Key: mapKey(config, partitionKey, sortKey),
    ConditionExpression: createIfNotExists
      ? undefined
      : `#${partitionKeyName} = :${partitionKeyName}${sortKeyName ? ` and #${sortKeyName} = :${sortKeyName}` : ''}`,
    ExpressionAttributeValues: marshall(
      mapAttributes(config, item, createIfNotExists).reduce(
        (attributeValues, [key, value]) => ({
          ...attributeValues,
          [`:${key}`]: value,
        }),
        {},
      ),
      {removeUndefinedValues: true},
    ),
    ExpressionAttributeNames: mapAttributes(config, item, createIfNotExists).reduce(
      (attributeNames, [key]) => ({
        ...attributeNames,
        [`#${key}`]: key,
      }),
      {},
    ),
    UpdateExpression: `SET ${Object.keys(item)
      .filter((key) => key !== config.partitionKeyName && key !== config.sortKeyName)
      .map((key) => `#${key}=:${key}`)
      .join(',')}`,
    ReturnValues: 'ALL_NEW',
  }
}

const mapAttributes = <Item extends object, PKName extends keyof Item, SKName extends keyof Item | never = never>(
  config: DDBReadOnlyClientConfiguration<Item, PKName, SKName>,
  item: Partial<Item>,
  createIfNotExists = false,
) => {
  return Object.entries(replaceUndefinedValuesWithNull(item)).filter(
    ([key]) =>
      !createIfNotExists || (createIfNotExists && key !== config.partitionKeyName && key !== config.sortKeyName),
  )
}
